/* Canale.java */

import java.util.Vector;

/**
 * Questa classe implementa il generico contenitore
 *  
 */
public class Canale extends Thread {
  private final static int MAX_EVENTS = 10;

  private Vector sottoscrittori = new Vector();

  private String[] msgInCoda = new String[MAX_EVENTS];

  private int msgInCodaLength = 0;

  public Canale() {
    super("Canale");
  }

  /**
   * Metodo eseguito dal processo Canale
   */
  public void run() {
    while (true) {
      try {
        Thread.sleep(2000);
        notifyMsg();
      }
      catch (Exception e) {
        System.out
            .println("" + this.getClass() + ", errore! Il seguente: " + e);
        e.printStackTrace();
      }

    }
  }

  public synchronized void addSubscriber(Sottoscrittore s) {
    //se il sottoscrittore non e' ancora stato registrato lo registro
    if (!sottoscrittori.contains(s))
      sottoscrittori.add(s);
  }

  public synchronized void removeSubscriber(Sottoscrittore s) {
    //se il sottoscrittore esiste lo de-sottoscrivo
    if (sottoscrittori.contains(s))
      sottoscrittori.removeElement(s);
  }

  /**
   * Metodo per la pubblicazione di messaggi da parte degli Editori.
   * 
   * Perche' questo metodo e' synchronized? Cosa accade ai Sottoscrittori
   * durante l'esecuzione di questo metodo?
   *  
   */
  public synchronized void put(String msg) {
    try {
      /*
       * Perche' c'e' un while e non un if? Quanti thread eseguono questo
       * metodo? Pensare a una possibile situazione di corsa critica se ci fosse
       * un if al posto del while.
       *  
       */
      while (msgInCodaLength > (msgInCoda.length - 1)) {
        wait();
      }
      msgInCoda[msgInCodaLength] = msg;
      msgInCodaLength++;
    }
    catch (Exception e) {
      System.out.println("Canale, si e' presentato il seguente errore: " + e);
      e.printStackTrace();
    }
  }

  /**
   * Metodo per la notifica dei messaggi ricevuti ai Sottoscrittori
   *  
   */
  private synchronized void notifyMsg() {
    /*
     * NOTA: durante l'esecuzione di questo metodo, i thread Editori non possono
     * eseguire il metodo put... Per molti sottoscrittori -> possibili effetti
     * di sequenzializzazione
     * 
     * Come risovere tali problemi?
     *  
     */
    String msg;
    int tot = sottoscrittori.size();

    /*
     * Recapito tutti i msg in coda a tutti i sottoscrittori.
     *  
     */
    for (int k = 0; k < msgInCodaLength; k++) {
      msg = msgInCoda[k];
      for (int i = 0; i < tot; i++) {
        Sottoscrittore sottoscrittoreIesimo = (Sottoscrittore) sottoscrittori
            .elementAt(i);
        //DEBUG
        //System.out.println("Canale, ecco il messaggio di cui faccio il push:
        // "+msg);
        sottoscrittoreIesimo.push(msg);
      }
    }
    // elimino i messaggi dalla coda
    msgInCodaLength = 0;
    
    /*
     * Sblocco processi eventualmente bloccati sulla put.
     * 
     * NOTA: qui e' necessaria la notifyAll perche' potrei avere diversi Thread
     * sospesi.
     */
    notifyAll();
  }
}